#include <qstringlist.h>
#include <qdatetime.h>
#include <qsocket.h>

#include "client.h"

Client::Client(const QString &host, Q_UINT16 port)
{
    /* create the socket and connect various of its signals */
    socket = new QSocket(this);
    QObject::connect(socket, SIGNAL(connected()),
		     this, SLOT(socketConnected()));
    QObject::connect(socket, SIGNAL(connectionClosed()),
		     this, SLOT(socketConnectionClosed()));
    QObject::connect(socket, SIGNAL(error(int)),
		     this, SLOT(socketError(int)));
    host_ = host;
    port_ = port;
    connected_ = false;
    status_ = -1;
    lock_ = false;
    commands_ = new QStringList();
}


Client::~Client()
{
}


void Client::run()
{
    QString data;
    bool even;

    even = true;
    while (true)
    {
	if (connected_)
	{
	    data = sendWait("status 1");
	    if (data.contains("OK Offline"))
	    {
		if (status_ >= 1)
		{
		    status_ = 0;
		    emit statusChanged(0, "Line 1 offline");
		}
	    }
	    if (data.contains("OK Online"))
	    {
		if (status_ <= 2)
		{
		    status_ = 3;
		    emit statusChanged(3, "Line 1 online");
		}
	    }
	    data = sendWait("status 2");
	    if (data.contains("OK Offline"))
	    {
		if (status_ == 4)
		{
		    status_ = 3;
		    emit statusChanged(3, "2nd Ch off");
		}
	    }
	    if (data.contains("OK Online"))
	    {
		if (status_ != 4)
		{
		    status_ = 4;
		    emit statusChanged(4, "2nd Ch on!");
		}
	    }
	    
	    if (even)
	    {
		data = sendWait("date");
		if (data.contains("OK"))
		{
		    emit setDate(data.remove(0, 3).stripWhiteSpace());
		}
		data = sendWait("route");
		if (data.contains("OK"))
		{
		    emit setRoute(data.remove(0, 3).toInt());
		}
		
		data = sendWait("time 1");
		if (data.contains("OK"))
		{
		    emit setCh1Time(data.remove(0, 3).stripWhiteSpace());
		}
		data = sendWait("chargetime 1");
		if (data.contains("OK"))
		{
		    /* TO DO:
		     * implement a way to hangup just a few seconds before
		     * the next charge time interval; should only be possible
		     * if longer than a certain time (1 minute?).
		     *
		     * Probably this is implemented in fli4l already?
		     *
		     * ch1ChargeTime_ = data.remove(0, 3).stripWhiteSpace();
		     */
		    emit setCh1ChargeTime(data.remove(0, 3).stripWhiteSpace());
		}
		data = sendWait("charge 1");
		if (data.contains("OK"))
		{
		    emit setCh1Charge(data.remove(0, 3).stripWhiteSpace());
		}
		data = sendWait("phone 1");
		if (data.contains("OK"))
		{
		    emit setCh1Phone(data.remove(0, 3).stripWhiteSpace());
		}
	    }
	    else
	    {
		data = sendWait("time 2");
		if (data.contains("OK"))
		{
		    emit setCh2Time(data.remove(0, 3).stripWhiteSpace());
		}
		data = sendWait("chargetime 2");
		if (data.contains("OK"))
		{
		    emit setCh2ChargeTime(data.remove(0, 3).stripWhiteSpace());
		}
		data = sendWait("charge 2");
		if (data.contains("OK"))
		{
		    emit setCh2Charge(data.remove(0, 3).stripWhiteSpace());
		}
		data = sendWait("phone 2");
		if (data.contains("OK"))
		{
		    emit setCh2Phone(data.remove(0, 3).stripWhiteSpace());
		}
	    }
	    even = !even;
	}
	msleep(1000);
    }
}


void Client::openConnection()
{
    status_ = -1;
    emit statusChanged(-1, "Try to connect to server");
    socket->connectToHost(host_, port_);
}


void Client::closeConnection()
{
    connected_ = false;
    socket->close();
    if (socket->state() == QSocket::Closing)
    {
	/* We have a delayed close. */
	connect(socket, SIGNAL(delayedCloseFinished()),
		SLOT(socketClosed()));
    } 
    else 
    {
	/* The socket is closed. */
	socketClosed();
    }
}


void Client::dial()
{
    if (sendWait("dial", true).contains("OK"))
    {
	if (status_ != 1)
	{
	    status_ = 1;
	    emit statusChanged(1, "Dialing...");
	}
    }
}


void Client::hangup()
{
    /*
     * This is not really the fine way - write
     * "hangup" to the socket no matter what
     * command is currently sent. Will mess
     * up the reply for other commands since
     * it does not care about synchronisation but
     * ensures that "hangup" is not delayed in
     * the while(lock_) loop of sendWait(QString).
     */
    socket->writeBlock("hangup\r\n", 8);
    socket->flush();
    if (sendWait("hangup", true).contains("OK"))
    {
	if (status_ != 2)
	{
	    status_ = 2;
	    emit statusChanged(2, "Hanging up...");
	}
    }
}


void Client::setNewRoute(int route)
{
    QString num;
    
    num.setNum(route);
    sendWait("route "+num);
}


QTime Client::createTimeObject(QString time)
{
    int h;
    int m;
    int s;

    h = time.left(2).toInt();
    m = time.mid(3, 4).toInt();
    s = time.right(2).toInt();
    
    return QTime(h, m, s);
}


QString Client::sendWait(QString msg)
{
    return sendWait(msg, true);
}


QString Client::sendWait(QString msg, bool noOut)
{
    QString data;
    data = "";
    int n;
//    int m;

    if (commands_->contains(msg))
    {
	if (! noOut)
	{
	    emit output(msg+" not send, already in queue!");
	}
	return data;
    }
    if (connected_)
    {
	commands_->append(msg);
    }
    else
    {
	if (! noOut)
	{
	    emit output("Not sent, not connected! "+msg);
	}
	return data;
    }

    /* FIX ME:
     * - possible dead lock
     * - check and set not simultaniously
     * (- should sleep for a random time)
     */
    while(lock_)
    {
	msleep(10);
	if (!connected_)
	{
	    if (! noOut)
	    {
		emit output("Not sent, not connected! "+msg);
	    }
	    return data;
	}    
    }
    lock_ = true;

    socket->writeBlock((msg+"\r\n").latin1(), msg.length()+2);
    socket->flush();
    commands_->remove(msg);
    n = 99;
    while (! socket->canReadLine())
    {
	socket->waitForMore(10);
	if (n <= 0)
	{
	    emit output("No reply from socket - waited 1 second. "+msg);
	    lock_ = false;
	    return data;
	}
	n--;
    }
    /* FIX ME:
     * Some commands may be delayed (like telmond-log-file).
     * Therefore we should wait until we get ERR or OK.
     * Make sure that we wait only until a timeout to avoid deadlocks.
     */
    while (socket->canReadLine())
    {
	data += socket->readLine();
    }
    lock_ = false;
    if (! noOut)
    {
	emit output(msg);
	emit output(data);
    }
    /* FIX ME:
     * For commands like help we need data
     * *before* the OK as well.
     *
     * only return data from this command
     * -> the last "OK" or "ERR" in our string.
     *
    n = data.find("OK", -1);
    m = data.find("ERR", -1);
    if (n>m)
    {
	data = data.mid(n);
    }
    else
    {
	if (m>0)
	{
	    data = data.mid(m);
	}
	}
    */

    return data;
}


QStringList Client::getCircuits()
{
    QStringList list;
    QString no;
    int n;
    int i;

    list.append("automatic");
    n = sendWait("circuits", true).remove(0, 3).toInt();
    for (i = 1; i <= n; i++)
    {
	no.setNum(i);
	list.append(
	    sendWait("circuit "+no, true).remove(0, 3).stripWhiteSpace());
    }
    return list;
}


void Client::socketConnected()
{
    connected_ = true;
    emit setCircuitList(getCircuits());
    status_ = 0;
    commands_->clear();
    emit statusChanged(0, "Connected to server");
}

void Client::socketConnectionClosed()
{
    connected_ = false;
    status_ = -1;
    emit statusChanged(-1, "Conn. closed by server");
}

void Client::socketClosed()
{
    connected_ = false;
    status_ = -1;
    emit statusChanged(-1, "Connection closed");
}

void Client::socketError(int e)
{
    QString s;

    connected_ = false;
    status_ = -1;
    s.setNum(e);
    emit statusChanged(-1, "Error number: "+s);
}
