#ifndef __client_h__
#define __client_h__

#include <qobject.h>
#include <qthread.h>

class QStringList;
class QSocket;
class QTime;

/**
 * Class client is responsible for all communication
 * with the fli4l server. It extends QThread that
 * it can to cyclic checks of the status.
 *
 * @author Roger Hartmann
 * @version "$Id$"
 */
class Client : public QObject, public QThread
{
    Q_OBJECT

public:

    /**
     * @param host Fli4l's host name or IP address.
     * @param port Fli4l's port (usually 5000).
     */
    Client(const QString &host, Q_UINT16 port);


    /**
     * Destructor. Currently not used.
     */
    ~Client();


    /**
     * QThread's run method. Responsible for cyclic status
     * checks that are responsible to emit most of the
     * signals.
     */ 
    void run();


    /**
     * Creates a QTime object out of a time string
     * as it is returned by the server (hh:mm:ss).
     * NO CHECKS for the format of the String!
     */
    QTime createTimeObject(QString time);


    /**
     * Send a message to the server. Calls
     * sendWait(msg, true).
     *
     * @param msg The message to send to Fli4l server.
     * @return The server's reply.
     */
    QString sendWait(QString msg); 


    /**
     * Send a message to the server. Only one message
     * can be sent at a time (method is synchronized) 
     * that we know which reply belongs to which msg.
     *
     * @param msg The message to send to Fli4l server.
     * @param noOut Set if you do not want ouput send
     *   by emitting signal output.
     * @return The server's reply.
     */
    QString sendWait(QString msg, bool noOut); 

    
    /**
     * @return The circuits (usually internet service
     * providers) - useful to create a QComboBox.
     */
    QStringList getCircuits();


public slots:
    /**
     * Closes the connection to the Fli4l Server.
     */
    void closeConnection();


    /**
     * Opens the connection to the Fli4l Server.
     */
    void openConnection();

    
    /**
     * Sends the "dial" command to the server and sets
     * the status accoardingly.
     *
     * @see status_
     */
    void dial();


    /**
     * Sends the "hangup" command to the server and sets
     * the status accoardingly.<br>
     * Note: The command will be sent directly (not using
     * sendWait(QString)) to trick the synchronisation of
     * that method. This is to ensure that our hangup
     * isn't kept in a waiting loop.
     *
     * @see status_
     */
    void hangup();

    
    /**
     * @param route set new route (usually ISP).
     *
     * @see getCircuits()
     */
    void setNewRoute(int route);


signals:
    /**
     * Emitted when the status was changed.
     * 
     * @param status The new status.
     * @param msg A descriptive message.
     * @see status_
     */
    void statusChanged(int status, QString msg);


    /**
     * The signal is used to write output (like
     * the server's replies).
     */
    void output(QString msg);


    /**
     * @param date The date of the server - to make use
     *   of in a user interface.
     */
    void setDate(QString date);


    /**
     * @param date The route (ISP) of the server - to make use
     *   of in a user interface.
     */
    void setRoute(int route);


    /**
     * @param date The circuit (ISP) list of the server
     *   - to make use of in a user interface.
     */
    void setCircuitList(QStringList list);


    /**
     * @param time Channel 1's online time.
     *   - to make use of in a user interface.
     */
    void setCh1Time(QString time);


    /**
     * @param chargeTime Channel 1's charge time.
     *   - to make use of in a user interface.
     */
    void setCh1ChargeTime(QString chargeTime);


    /**
     * @param charge Channel 1's charge.
     *   - to make use of in a user interface.
     */
    void setCh1Charge(QString charge);


    /**
     * @param phone Channel 1's phone number
     * when dialling or ISP name when connected.
     *   - to make use of in a user interface.
     */
    void setCh1Phone(QString phone);


    /**
     * @param time Channel 2's online time.
     *   - to make use of in a user interface.
     */
    void setCh2Time(QString time);


    /**
     * @param chargeTime Channel 2's charge time.
     *   - to make use of in a user interface.
     */
    void setCh2ChargeTime(QString chargeTime);


    /**
     * @param charge Channel 2's charge.
     *   - to make use of in a user interface.
     */
    void setCh2Charge(QString charge);


    /**
     * @param phone Channel 2's phone number
     * when dialling or ISP name when connected.
     *   - to make use of in a user interface.
     */
    void setCh2Phone(QString phone);


private slots:
    /**
     * Used internally - called when the socket is
     * connected to the server.
     * Emits the signal setCircuitList.
     */
    void socketConnected();


    /**
     * Used internally - called when the socket connection is
     * closed by the server.
     */
    void socketConnectionClosed();


    /**
     * Used internally - called when the socket is
     * closed.
     */
    void socketClosed();


    /**
     * Used internally - called when the socket emits
     * an error.
     */
    void socketError(int e);


private:
    /**
     * The socket that connects to the server.
     */
    QSocket *socket;


    /**
     * Contains the server's hostname.
     */
    QString host_;


    /**
     * Contains the port to connect to.
     */
    Q_UINT16 port_;


    /**
     * True if connected to the server.
     */
    bool connected_;


    /**
     * Contains the status of the server:
     * <p>
     * -1 Not connected to the server.<br>
     *  0 Connected to the server, offline.<br>
     *  1 Connected to the server, offline, dialling.<br>
     *  2 Connected to the server, online, hanging up.<br>
     *  3 Connected to the server, one line online.<br>
     *  4 Connected to the server, two lines online.<br>
     */
    int status_;


    /**
     * Used for synchronisation of the sendWait method.
     * Not really working that well - should use QMutex
     * instead when QT 3.0 is widely used.
     */
    bool lock_;


    /**
     * Contains a list of commands that is about to be sent
     * by method sendWait to ensure that one command does
     * not appear multiple times.
     */
    QStringList* commands_;
};
#endif
